/* named.h - include the local definitions in the right order
 * vix 28aug93 [original]
 *
 * $Id: named.h,v 1.1 1993/09/08 04:57:40 vixie Exp $
 */

#include "../conf/portability.h"
#include "../conf/options.h"

#include "pathnames.h"

#include "ns_defs.h"
#include "db_defs.h"

#include "ns_glob.h"
#include "db_glob.h"

#include "ns_func.h"
#include "db_func.h"
