#ifndef LINT
static char rcsid[] = "$Id: setitimer.c,v 1.4 1994/06/06 09:08:03 vixie Exp $";
#endif

/*
 * Setitimer emulation routine for UNICOS BIND.
 */
#if !defined(_CRAY)
int __bindcompat_setitimer;
#else
#include <sys/time.h>

int
__setitimer(int which, const struct itimerval *value,
	    struct itimerval *ovalue)
{
	if (alarm(value->it_value.tv_sec) >= 0)
		return (0);
	else
		return (-1);
}
#endif
